@@ -63,7 +63,6 @@ from google.cloud import bigquery
 from google.cloud import bigquery_v2
 from google.cloud.bigquery.dataset import Dataset
 from google.cloud.bigquery.dataset import DatasetReference
-from google.cloud.bigquery.schema import SchemaField
 from google.cloud.bigquery.table import Table
 from google.cloud._helpers import UTC
 from google.cloud.bigquery import dbapi, enums
@@ -154,7 +153,6 @@ class Config(object):
 
     CLIENT: Optional[bigquery.Client] = None
     CURSOR = None
-    DATASET = None
 
 
 def setUpModule():
@@ -164,9 +162,7 @@ def setUpModule():
 
 class TestBigQuery(unittest.TestCase):
     def setUp(self):
-        Config.DATASET = _make_dataset_id("bq_system_tests")
-        dataset = Config.CLIENT.create_dataset(Config.DATASET)
-        self.to_delete = [dataset]
+        self.to_delete = []
 
     def tearDown(self):
         policy_tag_client = PolicyTagManagerClient()
@@ -506,22 +502,6 @@ class TestBigQuery(unittest.TestCase):
         with self.assertRaises(exceptions.BadRequest):
             Config.CLIENT.delete_dataset(dataset)
 
-    def test_delete_job_metadata(self):
-        dataset_id = _make_dataset_id("us_east1")
-        self.temp_dataset(dataset_id, location="us-east1")
-        full_table_id = f"{Config.CLIENT.project}.{dataset_id}.test_delete_job_metadata"
-        table = Table(full_table_id, schema=[SchemaField("col", "STRING")])
-        Config.CLIENT.create_table(table)
-        query_job: bigquery.QueryJob = Config.CLIENT.query(
-            f"SELECT COUNT(*) FROM `{full_table_id}`", location="us-east1",
-        )
-        query_job.result()
-        self.assertIsNotNone(Config.CLIENT.get_job(query_job))
-
-        Config.CLIENT.delete_job_metadata(query_job)
-        with self.assertRaises(NotFound):
-            Config.CLIENT.get_job(query_job)
-
     def test_get_table_w_public_dataset(self):
         public = "bigquery-public-data"
         dataset_id = "samples"
@@ -1521,6 +1501,62 @@ class TestBigQuery(unittest.TestCase):
         self.assertGreater(stages_with_inputs, 0)
         self.assertGreater(len(plan), stages_with_inputs)
 
+    def test_dml_statistics(self):
+        table_schema = (
+            bigquery.SchemaField("foo", "STRING"),
+            bigquery.SchemaField("bar", "INTEGER"),
+        )
+
+        dataset_id = _make_dataset_id("bq_system_test")
+        self.temp_dataset(dataset_id)
+        table_id = "{}.{}.test_dml_statistics".format(Config.CLIENT.project, dataset_id)
+
+        # Create the table before loading so that the column order is deterministic.
+        table = helpers.retry_403(Config.CLIENT.create_table)(
+            Table(table_id, schema=table_schema)
+        )
+        self.to_delete.insert(0, table)
+
+        # Insert a few rows and check the stats.
+        sql = f"""
+            INSERT INTO `{table_id}`
+            VALUES ("one", 1), ("two", 2), ("three", 3), ("four", 4);
+        """
+        query_job = Config.CLIENT.query(sql)
+        query_job.result()
+
+        assert query_job.dml_stats is not None
+        assert query_job.dml_stats.inserted_row_count == 4
+        assert query_job.dml_stats.updated_row_count == 0
+        assert query_job.dml_stats.deleted_row_count == 0
+
+        # Update some of the rows.
+        sql = f"""
+            UPDATE `{table_id}`
+            SET bar = bar + 1
+            WHERE bar > 2;
+        """
+        query_job = Config.CLIENT.query(sql)
+        query_job.result()
+
+        assert query_job.dml_stats is not None
+        assert query_job.dml_stats.inserted_row_count == 0
+        assert query_job.dml_stats.updated_row_count == 2
+        assert query_job.dml_stats.deleted_row_count == 0
+
+        # Now delete a few rows and check the stats.
+        sql = f"""
+            DELETE FROM `{table_id}`
+            WHERE foo != "two";
+        """
+        query_job = Config.CLIENT.query(sql)
+        query_job.result()
+
+        assert query_job.dml_stats is not None
+        assert query_job.dml_stats.inserted_row_count == 0
+        assert query_job.dml_stats.updated_row_count == 0
+        assert query_job.dml_stats.deleted_row_count == 3
+
     def test_dbapi_w_standard_sql_types(self):
         for sql, expected in helpers.STANDARD_SQL_EXAMPLES:
             Config.CURSOR.execute(sql)
@@ -1566,20 +1602,6 @@ class TestBigQuery(unittest.TestCase):
         row_tuples = [r.values() for r in rows]
         self.assertEqual(row_tuples, [(5, "foo"), (6, "bar"), (7, "baz")])
 
-    def test_dbapi_create_view(self):
-
-        query = """
-        CREATE VIEW {}.dbapi_create_view
-        AS SELECT name, SUM(number) AS total
-        FROM `bigquery-public-data.usa_names.usa_1910_2013`
-        GROUP BY name;
-        """.format(
-            Config.DATASET
-        )
-
-        Config.CURSOR.execute(query)
-        self.assertEqual(Config.CURSOR.rowcount, 0, "expected 0 rows")
-
     @unittest.skipIf(
         bigquery_storage is None, "Requires `google-cloud-bigquery-storage`"
     )
@@ -2172,6 +2194,85 @@ class TestBigQuery(unittest.TestCase):
         assert len(rows) == 1
         assert rows[0].max_value == 100.0
 
+    def test_create_tvf_routine(self):
+        from google.cloud.bigquery import Routine, RoutineArgument, RoutineType
+
+        StandardSqlDataType = bigquery_v2.types.StandardSqlDataType
+        StandardSqlField = bigquery_v2.types.StandardSqlField
+        StandardSqlTableType = bigquery_v2.types.StandardSqlTableType
+
+        INT64 = StandardSqlDataType.TypeKind.INT64
+        STRING = StandardSqlDataType.TypeKind.STRING
+
+        client = Config.CLIENT
+
+        dataset = self.temp_dataset(_make_dataset_id("create_tvf_routine"))
+        routine_ref = dataset.routine("test_tvf_routine")
+
+        routine_body = """
+            SELECT int_col, str_col
+            FROM (
+                UNNEST([1, 2, 3]) int_col
+                JOIN
+                (SELECT str_col FROM UNNEST(["one", "two", "three"]) str_col)
+                ON TRUE
+            )
+            WHERE int_col > threshold
+            """
+
+        return_table_type = StandardSqlTableType(
+            columns=[
+                StandardSqlField(
+                    name="int_col", type=StandardSqlDataType(type_kind=INT64),
+                ),
+                StandardSqlField(
+                    name="str_col", type=StandardSqlDataType(type_kind=STRING),
+                ),
+            ]
+        )
+
+        routine_args = [
+            RoutineArgument(
+                name="threshold", data_type=StandardSqlDataType(type_kind=INT64),
+            )
+        ]
+
+        routine_def = Routine(
+            routine_ref,
+            type_=RoutineType.TABLE_VALUED_FUNCTION,
+            arguments=routine_args,
+            return_table_type=return_table_type,
+            body=routine_body,
+        )
+
+        # Create TVF routine.
+        client.delete_routine(routine_ref, not_found_ok=True)
+        routine = client.create_routine(routine_def)
+
+        assert routine.body == routine_body
+        assert routine.return_table_type == return_table_type
+        assert routine.arguments == routine_args
+
+        # Execute the routine to see if it's working as expected.
+        query_job = client.query(
+            f"""
+            SELECT int_col, str_col
+            FROM `{routine.reference}`(1)
+            ORDER BY int_col, str_col ASC
+            """
+        )
+
+        result_rows = [tuple(row) for row in query_job.result()]
+        expected = [
+            (2, "one"),
+            (2, "three"),
+            (2, "two"),
+            (3, "one"),
+            (3, "three"),
+            (3, "two"),
+        ]
+        assert result_rows == expected
+
     def test_create_table_rows_fetch_nested_schema(self):
         table_name = "test_table"
         dataset = self.temp_dataset(_make_dataset_id("create_table_nested_schema"))
@@ -2341,104 +2442,6 @@ class TestBigQuery(unittest.TestCase):
         page = next(pages)
         self.assertEqual(page.num_items, num_last_page)
 
-    def test_parameterized_types_round_trip(self):
-        client = Config.CLIENT
-        table_id = f"{Config.DATASET}.test_parameterized_types_round_trip"
-        fields = (
-            ("n", "NUMERIC"),
-            ("n9", "NUMERIC(9)"),
-            ("n92", "NUMERIC(9, 2)"),
-            ("bn", "BIGNUMERIC"),
-            ("bn9", "BIGNUMERIC(38)"),
-            ("bn92", "BIGNUMERIC(38, 22)"),
-            ("s", "STRING"),
-            ("s9", "STRING(9)"),
-            ("b", "BYTES"),
-            ("b9", "BYTES(9)"),
-        )
-        self.to_delete.insert(0, Table(f"{client.project}.{table_id}"))
-        client.query(
-            "create table {} ({})".format(
-                table_id, ", ".join(" ".join(f) for f in fields)
-            )
-        ).result()
-        table = client.get_table(table_id)
-        table_id2 = table_id + "2"
-        self.to_delete.insert(0, Table(f"{client.project}.{table_id2}"))
-        client.create_table(Table(f"{client.project}.{table_id2}", table.schema))
-        table2 = client.get_table(table_id2)
-
-        self.assertEqual(tuple(s._key()[:2] for s in table2.schema), fields)
-
-    def test_table_snapshots(self):
-        from google.cloud.bigquery import CopyJobConfig
-        from google.cloud.bigquery import OperationType
-
-        client = Config.CLIENT
-
-        source_table_path = f"{client.project}.{Config.DATASET}.test_table"
-        snapshot_table_path = f"{source_table_path}_snapshot"
-
-        # Create the table before loading so that the column order is predictable.
-        schema = [
-            bigquery.SchemaField("foo", "INTEGER"),
-            bigquery.SchemaField("bar", "STRING"),
-        ]
-        source_table = helpers.retry_403(Config.CLIENT.create_table)(
-            Table(source_table_path, schema=schema)
-        )
-        self.to_delete.insert(0, source_table)
-
-        # Populate the table with initial data.
-        rows = [{"foo": 1, "bar": "one"}, {"foo": 2, "bar": "two"}]
-        load_job = Config.CLIENT.load_table_from_json(rows, source_table)
-        load_job.result()
-
-        # Now create a snapshot before modifying the original table data.
-        copy_config = CopyJobConfig()
-        copy_config.operation_type = OperationType.SNAPSHOT
-
-        copy_job = client.copy_table(
-            sources=source_table_path,
-            destination=snapshot_table_path,
-            job_config=copy_config,
-        )
-        copy_job.result()
-
-        snapshot_table = client.get_table(snapshot_table_path)
-        self.to_delete.insert(0, snapshot_table)
-
-        # Modify data in original table.
-        sql = f'INSERT INTO `{source_table_path}`(foo, bar) VALUES (3, "three")'
-        query_job = client.query(sql)
-        query_job.result()
-
-        # List rows from the source table and compare them to rows from the snapshot.
-        rows_iter = client.list_rows(source_table_path)
-        rows = sorted(row.values() for row in rows_iter)
-        assert rows == [(1, "one"), (2, "two"), (3, "three")]
-
-        rows_iter = client.list_rows(snapshot_table_path)
-        rows = sorted(row.values() for row in rows_iter)
-        assert rows == [(1, "one"), (2, "two")]
-
-        # Now restore the table from the snapshot and it should again contain the old
-        # set of rows.
-        copy_config = CopyJobConfig()
-        copy_config.operation_type = OperationType.RESTORE
-        copy_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
-
-        copy_job = client.copy_table(
-            sources=snapshot_table_path,
-            destination=source_table_path,
-            job_config=copy_config,
-        )
-        copy_job.result()
-
-        rows_iter = client.list_rows(source_table_path)
-        rows = sorted(row.values() for row in rows_iter)
-        assert rows == [(1, "one"), (2, "two")]
-
     def temp_dataset(self, dataset_id, location=None):
         project = Config.CLIENT.project
         dataset_ref = bigquery.DatasetReference(project, dataset_id)
@@ -2469,3 +2472,108 @@ def _table_exists(t):
         return True
     except NotFound:
         return False
+
+
+def test_dbapi_create_view(dataset_id):
+
+    query = f"""
+    CREATE VIEW {dataset_id}.dbapi_create_view
+    AS SELECT name, SUM(number) AS total
+    FROM `bigquery-public-data.usa_names.usa_1910_2013`
+    GROUP BY name;
+    """
+
+    Config.CURSOR.execute(query)
+    assert Config.CURSOR.rowcount == 0, "expected 0 rows"
+
+
+def test_parameterized_types_round_trip(dataset_id):
+    client = Config.CLIENT
+    table_id = f"{dataset_id}.test_parameterized_types_round_trip"
+    fields = (
+        ("n", "NUMERIC"),
+        ("n9", "NUMERIC(9)"),
+        ("n92", "NUMERIC(9, 2)"),
+        ("bn", "BIGNUMERIC"),
+        ("bn9", "BIGNUMERIC(38)"),
+        ("bn92", "BIGNUMERIC(38, 22)"),
+        ("s", "STRING"),
+        ("s9", "STRING(9)"),
+        ("b", "BYTES"),
+        ("b9", "BYTES(9)"),
+    )
+    client.query(
+        "create table {} ({})".format(table_id, ", ".join(" ".join(f) for f in fields))
+    ).result()
+    table = client.get_table(table_id)
+    table_id2 = table_id + "2"
+    client.create_table(Table(f"{client.project}.{table_id2}", table.schema))
+    table2 = client.get_table(table_id2)
+
+    assert tuple(s._key()[:2] for s in table2.schema) == fields
+
+
+def test_table_snapshots(dataset_id):
+    from google.cloud.bigquery import CopyJobConfig
+    from google.cloud.bigquery import OperationType
+
+    client = Config.CLIENT
+
+    source_table_path = f"{client.project}.{dataset_id}.test_table"
+    snapshot_table_path = f"{source_table_path}_snapshot"
+
+    # Create the table before loading so that the column order is predictable.
+    schema = [
+        bigquery.SchemaField("foo", "INTEGER"),
+        bigquery.SchemaField("bar", "STRING"),
+    ]
+    source_table = helpers.retry_403(Config.CLIENT.create_table)(
+        Table(source_table_path, schema=schema)
+    )
+
+    # Populate the table with initial data.
+    rows = [{"foo": 1, "bar": "one"}, {"foo": 2, "bar": "two"}]
+    load_job = Config.CLIENT.load_table_from_json(rows, source_table)
+    load_job.result()
+
+    # Now create a snapshot before modifying the original table data.
+    copy_config = CopyJobConfig()
+    copy_config.operation_type = OperationType.SNAPSHOT
+
+    copy_job = client.copy_table(
+        sources=source_table_path,
+        destination=snapshot_table_path,
+        job_config=copy_config,
+    )
+    copy_job.result()
+
+    # Modify data in original table.
+    sql = f'INSERT INTO `{source_table_path}`(foo, bar) VALUES (3, "three")'
+    query_job = client.query(sql)
+    query_job.result()
+
+    # List rows from the source table and compare them to rows from the snapshot.
+    rows_iter = client.list_rows(source_table_path)
+    rows = sorted(row.values() for row in rows_iter)
+    assert rows == [(1, "one"), (2, "two"), (3, "three")]
+
+    rows_iter = client.list_rows(snapshot_table_path)
+    rows = sorted(row.values() for row in rows_iter)
+    assert rows == [(1, "one"), (2, "two")]
+
+    # Now restore the table from the snapshot and it should again contain the old
+    # set of rows.
+    copy_config = CopyJobConfig()
+    copy_config.operation_type = OperationType.RESTORE
+    copy_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
+
+    copy_job = client.copy_table(
+        sources=snapshot_table_path,
+        destination=source_table_path,
+        job_config=copy_config,
+    )
+    copy_job.result()
+
+    rows_iter = client.list_rows(source_table_path)
+    rows = sorted(row.values() for row in rows_iter)
+    assert rows == [(1, "one"), (2, "two")]
